/*

  gbepshow (ver. 6.4) -- Show the reduced likelihood of power exponential distribution
  
  (C) 2022 Giulio Bottazzi

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  (version 2) as published by the Free Software Foundation;
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#include "tools.h"
/* #include "multimin.h" */
#include "assert.h"

#include <gsl/gsl_sf_gamma.h>
/* #include <gsl/gsl_roots.h> */


/* Global Variables */
/* ---------------- */

/* store data */
double *Data; /*the array of data*/
size_t Size;/*the number of data*/

/* output management */
FILE *Fmessages;/*stream for computation messages*/
FILE *Fparams;/*strem for final parameters*/
FILE *Ffitted;/*strem for distribution or density*/

/*------------------*/


/* "reduced" average negative log-likelihood: the estimate of a has
   been already substituted in the likelihood expression */
double loglikered(const double b, const double mu){

  double sum=0.0;
  unsigned utmp1;

  for(utmp1=0;utmp1<Size;utmp1++){
    const double dtmp1=fabs(mu-Data[utmp1]);
    if(dtmp1>0) sum+=pow(fabs(mu-Data[utmp1]),b);
  }
  
  return(log(2)+log(sum/Size)/b + log(b)/b+gsl_sf_lngamma(1./b+1.) + 1./b);
}

/* "reduced" average negative log-likelihood with position parameter m
   interpolated between observations */
double loglikered_spec(const double b, const double mu){

  if(b>1) {
    return(loglikered(b,mu));
  }
  else {
    int i;
    for(i=0;Data[i]<=mu && i<=Size;i++);
/*     fprintf(stderr,"%e %e %e\n",Data[i-1],mu,Data[i]); */
    if(i==0){
      return(loglikered(b,Data[0]));
    }
    else if(i==Size){
      return(loglikered(b,Data[Size-1]));
    }
    else{
      return((loglikered(b,Data[i-1])*(Data[i]-mu)+
	     loglikered(b,Data[i])*(mu-Data[i-1]))/(Data[i]-Data[i-1]));
    }
  }

}

/* the average of the exponent expression (x-m)^b */
double onlysum(const double b, const double mu){

  double sum=0.0;
  unsigned utmp1;

  for(utmp1=0;utmp1<Size;utmp1++){
    const double dtmp1=fabs(mu-Data[utmp1]);
    if(dtmp1>0) sum+=pow(fabs(mu-Data[utmp1]),b);
  }
  
  return(sum/Size);
}



int main(int argc,char* argv[]){
  
  /* varia */
  int i,j;

  /* set various default options */
  int N_b = 10;
  int N_mu =10;
  double mumin=-1.;
  double mumax=1.;
  double bmin=.5;
  double bmax=1.5;
  double bstep,mustep;
  unsigned O_output=0;

  int outype=0;
  double (* outfun)(double,double) ;

  char *splitstring = strdup(" \t");

  /* COMMAND LINE PROCESSING */
  /* ----------------------- */ 
  int opt;
    
  /* read the command line */    
  while((opt=getopt_long(argc,argv,"M:B:m:b:hF:",gb_long_options, &gb_option_index))!=EOF){
    if(opt==0){
      gbutils_header(argv[0],stdout);
      exit(0);
    }
    else if(opt=='?'){
      fprintf(stderr,"option %c not recognized\n",optopt);
      return(-1);
    }
    else if(opt=='h'){
      /*print short help*/
      fprintf(stdout,"Plot the reduced negative average log-likelihood of the power exponential distribution\n");
      fprintf(stdout,"on a set of data as a function of m and b. The ouput is made of two blocks. In the\n");
      fprintf(stdout,"first block the values for m and b are computed on a regular grid. In the second block\n");
      fprintf(stdout,"the observations provided are used as value for m. Output is thought to be sent to gnuplot\n");
      fprintf(stdout,"\nUsage: %s [options] [files]\n\n",argv[0]);
      fprintf(stdout,"Options:  \n");
      fprintf(stdout," -O  output type (default 0)   \n");
      fprintf(stdout,"      0  reduced average negative log-likelihood: the estimate of a has\n");
      fprintf(stdout,"         been already substituted in the likelihood expression\n");
      fprintf(stdout,"      1  reduced average negative log-likelihood with position parameter m\n");
      fprintf(stdout,"         interpolated between observations\n");
      fprintf(stdout,"      2  the average of the exponent expression (x-m)^b\n");
      fprintf(stdout," -M set the number of m values (default 10)\n");
      fprintf(stdout," -B set the number of b values (default 10)\n");
      fprintf(stdout," -m set the range of m with colon separated values min:max (default -1.:1)\n");
      fprintf(stdout," -b set the range of b with colon separated values min:max (default .5:1.5)\n");
      fprintf(stdout," -F  input fields separators (default \" \\t\")\n");
      fprintf(stdout," -h this help\n");
      fprintf(stdout,"Examples:\n");
      fprintf(stdout," splot '<gbepshow file' index 0 w l  to plot the log-likelihood of the observations\n");
      fprintf(stdout,"                                     in 'file' on a regular grid from inside gnuplot\n");
      fprintf(stdout," gbepshow file | gbplot splot 'index 0 w l'  Using the gbplot command\n");
      exit(0);
    }
    else if(opt=='F'){
      /*set the fields separator string*/
      free(splitstring);
      splitstring = strdup(optarg);
    }
    else if(opt=='O'){
      O_output = atoi(optarg);
    }
    else if(opt=='B'){
      N_b = atoi(optarg);
    }
    else if(opt=='b'){
      if(!strchr (optarg,':')){
	bmin=bmax=atof(optarg);
      }
      else{
	char *stmp1=strdup (optarg);
	bmin = atof(strtok (stmp1,":"));
	bmax= atof(strtok (NULL,":"));
	free(stmp1);
      }
    }
    else if(opt=='M'){
      N_mu = atoi(optarg);
    }
    else if(opt=='m'){
      if(!strchr (optarg,':')){
	mumin=mumax=atof(optarg);
      }
      else{
	char *stmp1=strdup (optarg);
	mumin = atof(strtok (stmp1,":"));
	mumax= atof(strtok (NULL,":"));
	free(stmp1);
      }
    }
  }
   
  /* initialize global variables */
  initialize_program(argv[0]);

  /* customized admin. of errors */
  /* --------------------------- */
  gsl_set_error_handler_off ();

  
  /* load Data */
  /* --------- */

  load(&Data,&Size,0,splitstring);

  if (Size <= 1) {
    fprintf(Fmessages,"the number of onservations must be at least 2\n");
    exit(-1);
  }

  /* sort data */
  /* --------- */
  qsort(Data,Size,sizeof(double),sort_by_value);


  /* chose the function to use */
  /* ------------------------- */

  switch(O_output){
  case 0:
    outfun=loglikered;
    break;
  case 1:
    outfun=loglikered_spec;
    break;
  case 2:
    outfun=onlysum;
    break;
  default:
    outfun=loglikered;
  }

  /* "reduced" maximum likelyhood plot */
  /* --------------------------------- */

  bstep = (bmax-bmin)/N_b;
  mustep = (mumax-mumin)/N_mu;


  for(j=0;j<N_mu;j++){
    const double mu=mumin+mustep*j;
    for(i=0;i<N_b;i++){
      const double b=bmin+bstep*i;
      if(N_mu>1) printf("%e ",mu);
      if(N_b>1) printf("%e ",b);
      printf("%e\n",(*outfun)(b,mu));
    }
    if(N_b>1) printf("\n");
  }



/*   for(i=0;i<N_b;i++){ */
/*     const double b=bmin+bstep*i; */
/*     for(j=0;j<N_mu;j++){ */
/*       const double mu=mumin+mustep*j; */
/*       if(N_b>1) printf("%f ",b); */
/*       if(N_mu>1) printf("%f ",mu); */
/*       printf("%f\n",(*outfun)(b,mu)); */
/*     } */
/*     if(N_mu>1) printf("\n"); */
/*   } */

  printf("\n\n");

  for(j=0;j<Size;j++){
    if( Data[j]<=mumax && Data[j]>=mumin){
      for(i=0;i<N_b;i++){
	const double b=bmin+bstep*i;
	printf("%e ",Data[j]);
	if(N_b>1) printf("%e ",b);
	printf("%e\n",(*outfun)(b,Data[j]));
      }
      if(N_b>1) printf("\n");
    }
  }

/*   for(i=0;i<N_b;i++){ */
/*     const double b=bmin+bstep*i; */
/*     for(j=0;j<Size;j++){ */
/*       if( Data[j]<=mumax && Data[j]>=mumin){ */
/* 	if(N_b>1) printf("%f ",b); */
/* 	printf("%f %f\n",Data[j],(*outfun)(b,Data[j])); */
/*       } */
/*     } */
/*     printf("\n"); */
/*   } */



/*   printf("\n \n"); */

/*   for(i=0;i<N_b;i++){ */
/*     const double b=bmin+bstep*i; */
/*     for(j=0;j<Size;j++){ */
/*       if(Data[j]>mumin && Data[j]<mumax){ */
/* 	const double dtmp1=Data[j]; */
/* 	const double dtmp2=(Data[j]+Data[j+1])*.5; */
/* 	printf("%f %f %f %f\n",b,dtmp1, */
/* 	       loglikered(b,dtmp1),loglikered_spec(b,dtmp1)); */
/* 	printf("%f %f %f %f\n",b,dtmp2, */
/* 	       loglikered(b,dtmp2),loglikered_spec(b,dtmp2)); */
/*       } */
/*     } */
/*     printf("\n"); */
/*   } */

  free(Data);
  free(splitstring);
  exit(0);

}
